package dev.pfaff.glass.glfw;

import dev.pfaff.glass.glfw.callbacks.ErrorCallback;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.lang.foreign.*;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.nio.charset.StandardCharsets;

import static java.lang.foreign.ValueLayout.*;
import static java.lang.invoke.MethodType.methodType;

public final class GLFW {
	static final Linker LINKER = Linker.nativeLinker();

	private static final MethodHandle MH_errorCallbackImpl;
	public static final FunctionDescriptor FD_GLFWerrorfun = FunctionDescriptor.ofVoid(JAVA_INT,
			ADDRESS.withTargetLayout(JAVA_BYTE));

	private static Arena errorCallbackArena;

	static {
		try {
			MH_errorCallbackImpl = MethodHandles.lookup().findStatic(GLFW.class,
					"errorCallbackImpl",
					methodType(void.class, ErrorCallback.class, int.class, MemorySegment.class));
		} catch (NoSuchMethodException | IllegalAccessException e) {
			throw new RuntimeException(e);
		}
	}

	private static void errorCallbackImpl(ErrorCallback callback, int code, MemorySegment descriptionPtr) {
		var description = descriptionPtr.reinterpret(Long.MAX_VALUE).getString(0, StandardCharsets.UTF_8);
		callback.call(code, description);
	}

	public static final AddressLayout ML_GLFWwindow = ADDRESS.withName("GLFWwindow*");
	public static final AddressLayout ML_GLFWmonitor = ADDRESS.withName("GLFWmonitor*");
	public static final AddressLayout ML_GLFWerrorfun = ADDRESS.withName("GLFWerrorfun");

	public static final MethodHandle MH_glfwInit = LINKER.downcallHandle(FunctionDescriptor.of(JAVA_BOOLEAN));
	public static final MethodHandle MH_glfwTerminate = LINKER.downcallHandle(FunctionDescriptor.ofVoid());
	public static final MethodHandle MH_glfwPollEvents = LINKER.downcallHandle(FunctionDescriptor.ofVoid());
	public static final MethodHandle MH_glfwSetErrorCallback = LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWerrorfun,
			ML_GLFWerrorfun));

	private final MemorySegment MS_glfwInit;
	private final MemorySegment MS_glfwTerminate;
	private final MemorySegment MS_glfwPollEvents;
	private final MemorySegment MS_glfwSetErrorCallback;

	final MemorySegment MS_glfwMakeContextCurrent;
	final MemorySegment MS_glfwGetCurrentContext;
	final MemorySegment MS_glfwSwapInterval;
	final MemorySegment MS_glfwExtensionSupported;
	final MemorySegment MS_glfwCreateWindow;
	final MemorySegment MS_glfwWindowHint;
	final MemorySegment MS_glfwDestroyWindow;
	final MemorySegment MS_glfwWindowShouldClose;
	final MemorySegment MS_glfwGetFramebufferSize;
	final MemorySegment MS_glfwGetWindowSize;
	final MemorySegment MS_glfwGetWindowContentScale;
	final MemorySegment MS_glfwGetWindowAttrib;
	final MemorySegment MS_glfwSwapBuffers;
	final MemorySegment MS_glfwShowWindow;
	final MemorySegment MS_glfwHideWindow;
	final MemorySegment MS_glfwSetWindowSize;

	final MemorySegment MS_glfwSetCharCallback;
	final MemorySegment MS_glfwSetCursorPosCallback;
	final MemorySegment MS_glfwSetKeyCallback;
	final MemorySegment MS_glfwSetMouseButtonCallback;
	final MemorySegment MS_glfwSetDropCallback;
	final MemorySegment MS_glfwSetScrollCallback;
	final MemorySegment MS_glfwSetWindowPosCallback;
	final MemorySegment MS_glfwSetWindowSizeCallback;
	final MemorySegment MS_glfwSetWindowCloseCallback;
	final MemorySegment MS_glfwSetWindowRefreshCallback;
	final MemorySegment MS_glfwSetWindowFocusCallback;
	final MemorySegment MS_glfwSetWindowIconifyCallback;
	final MemorySegment MS_glfwSetWindowMaximizeCallback;
	final MemorySegment MS_glfwSetFramebufferSizeCallback;
	final MemorySegment MS_glfwSetWindowContentScaleCallback;

	public GLFW(SymbolLookup symbolLookup) {
		this.MS_glfwInit = symbolLookup.findOrThrow("glfwInit");
		this.MS_glfwTerminate = symbolLookup.findOrThrow("glfwTerminate");
		this.MS_glfwPollEvents = symbolLookup.findOrThrow("glfwPollEvents");
		this.MS_glfwSetErrorCallback = symbolLookup.findOrThrow("glfwSetErrorCallback");

		this.MS_glfwMakeContextCurrent = symbolLookup.findOrThrow("glfwMakeContextCurrent");
		this.MS_glfwGetCurrentContext = symbolLookup.findOrThrow("glfwGetCurrentContext");
		this.MS_glfwSwapInterval = symbolLookup.findOrThrow("glfwSwapInterval");
		this.MS_glfwExtensionSupported = symbolLookup.findOrThrow("glfwExtensionSupported");
		this.MS_glfwCreateWindow = symbolLookup.findOrThrow("glfwCreateWindow");
		this.MS_glfwWindowHint = symbolLookup.findOrThrow("glfwWindowHint");
		this.MS_glfwDestroyWindow = symbolLookup.findOrThrow("glfwDestroyWindow");
		this.MS_glfwWindowShouldClose = symbolLookup.findOrThrow("glfwWindowShouldClose");
		this.MS_glfwGetFramebufferSize = symbolLookup.findOrThrow("glfwGetFramebufferSize");
		this.MS_glfwGetWindowSize = symbolLookup.findOrThrow("glfwGetWindowSize");
		this.MS_glfwGetWindowContentScale = symbolLookup.findOrThrow("glfwGetWindowContentScale");
		this.MS_glfwGetWindowAttrib = symbolLookup.findOrThrow("glfwGetWindowAttrib");
		this.MS_glfwSwapBuffers = symbolLookup.findOrThrow("glfwSwapBuffers");
		this.MS_glfwShowWindow = symbolLookup.findOrThrow("glfwShowWindow");
		this.MS_glfwHideWindow = symbolLookup.findOrThrow("glfwHideWindow");
		this.MS_glfwSetWindowSize = symbolLookup.findOrThrow("glfwSetWindowSize");

		this.MS_glfwSetCharCallback = symbolLookup.findOrThrow("glfwSetCharCallback");
		this.MS_glfwSetCursorPosCallback = symbolLookup.findOrThrow("glfwSetCursorPosCallback");
		this.MS_glfwSetKeyCallback = symbolLookup.findOrThrow("glfwSetKeyCallback");
		this.MS_glfwSetMouseButtonCallback = symbolLookup.findOrThrow("glfwSetMouseButtonCallback");
		this.MS_glfwSetDropCallback = symbolLookup.findOrThrow("glfwSetDropCallback");
		this.MS_glfwSetScrollCallback = symbolLookup.findOrThrow("glfwSetScrollCallback");
		this.MS_glfwSetWindowPosCallback = symbolLookup.findOrThrow("glfwSetWindowPosCallback");
		this.MS_glfwSetWindowSizeCallback = symbolLookup.findOrThrow("glfwSetWindowSizeCallback");
		this.MS_glfwSetWindowCloseCallback = symbolLookup.findOrThrow("glfwSetWindowCloseCallback");
		this.MS_glfwSetWindowRefreshCallback = symbolLookup.findOrThrow("glfwSetWindowRefreshCallback");
		this.MS_glfwSetWindowFocusCallback = symbolLookup.findOrThrow("glfwSetWindowFocusCallback");
		this.MS_glfwSetWindowIconifyCallback = symbolLookup.findOrThrow("glfwSetWindowIconifyCallback");
		this.MS_glfwSetWindowMaximizeCallback = symbolLookup.findOrThrow("glfwSetWindowMaximizeCallback");
		this.MS_glfwSetFramebufferSizeCallback = symbolLookup.findOrThrow("glfwSetFramebufferSizeCallback");
		this.MS_glfwSetWindowContentScaleCallback = symbolLookup.findOrThrow("glfwSetWindowContentScaleCallback");
	}

	// TODO: error callback

	/**
	 * Initializes GLFW.
	 */
	public void init() {
		boolean result;
		try {
			result = (boolean) MH_glfwInit.invokeExact(MS_glfwInit);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
		if (!result) {
			throw new RuntimeException("Failed to initialize GLFW");
		}
	}

	/**
	 * Deinitializes GLFW. You must call {@link #init()} after this call to use the library again.
	 */
	public void terminate() {
		try {
			MH_glfwTerminate.invokeExact(MS_glfwTerminate);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void pollEvents() {
		try {
			MH_glfwPollEvents.invokeExact(MS_glfwPollEvents);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	static <T> Arena setCallback(@NonNull MethodHandle MH_setCallback,
								 @NonNull MemorySegment MS_setCallback,
								 @NonNull T callback,
								 @NonNull MethodHandle invokerStub,
								 @NonNull FunctionDescriptor invokerFunctionDescriptor,
								 @Nullable Arena oldArena) {
		Arena newArena = Arena.ofShared();
		try {
			var stub = LINKER.upcallStub(invokerStub.bindTo(callback), invokerFunctionDescriptor, newArena);
			var _ = (MemorySegment) MH_setCallback.invokeExact(MS_setCallback, stub);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
		if (oldArena != null) oldArena.close();
		return newArena;
	}

	public void setErrorCallback(ErrorCallback callback) {
		errorCallbackArena = setCallback(MH_glfwSetErrorCallback,
				MS_glfwSetErrorCallback,
				callback,
				MH_errorCallbackImpl,
				FD_GLFWerrorfun,
				errorCallbackArena);
	}
}
